"""General plotting utilities."""

from __future__ import annotations

__all__ = [
    "figwh",
    "autoscale_off",
    "autoscale_to",
    "LabeledCursor",
    "place_inset",
    "plot_array",
    "plot_array_2d",
    "gradient_fill",
    "plot_slices",
    "fermiline",
]

import contextlib
import copy
from typing import TYPE_CHECKING, Literal

import matplotlib
import matplotlib.colors as mcolors
import matplotlib.image
import matplotlib.patches
import matplotlib.path
import matplotlib.pyplot as plt
import numpy as np
import numpy.typing as npt
import xarray as xr
from matplotlib.widgets import AxesWidget

from erlab.plotting.annotations import fancy_labels, label_subplot_properties
from erlab.plotting.colors import (
    InversePowerNorm,
    _ez_inset,
    axes_textcolor,
    gen_2d_colormap,
    nice_colorbar,
)

if TYPE_CHECKING:
    from collections.abc import Callable, Iterable, Sequence

figure_width_ref = {
    "aps": [3.4, 7.0],
    "aip": [3.37, 6.69],
    "nature": [88 / 25.4, 180 / 25.4],
}


def figwh(ratio=0.6180339887498948, wide=0, wscale=1, style="aps", fixed_height=True):
    if isinstance(ratio, str):
        ratio = float(ratio) * 2 / (1 + np.sqrt(5))
    w = figure_width_ref[style][wide]
    if fixed_height:
        h = w * ratio
    else:
        h = w * wscale * ratio
    return w * wscale, h


@contextlib.contextmanager
def autoscale_off(ax: matplotlib.axes.Axes | None = None):
    if ax is None:
        ax = plt.gca()
    xauto, yauto = ax.get_autoscalex_on(), ax.get_autoscaley_on()
    xl, yl = ax.get_xlim(), ax.get_ylim()
    ax.set_xlim(*xl)
    ax.set_ylim(*yl)
    try:
        yield
    finally:
        ax.autoscale(enable=xauto, axis="x")
        ax.autoscale(enable=yauto, axis="y")


def autoscale_to(arr, margin=0.2):
    mn, mx = min(arr), max(arr)
    diff = margin * (mx - mn)
    return mn - diff, mx + diff


class LabeledCursor(AxesWidget):
    r"""A crosshair cursor that spans the axes and moves with mouse cursor.

    For the cursor to remain responsive you must keep a reference to it.
    Unlike `matplotlib.widgets.Cursor`, this also shows the current
    cursor location.

    Parameters
    ----------
    ax
        The `matplotlib.axes.Axes` to attach the cursor to.
    horizOn
        Whether to draw the horizontal line.
    vertOn
        Whether to draw the vertical line.
    textOn
        Whether to show current cursor location.
    useblit
        Use blitting for faster drawing if supported by the backend.
    textprops
        Keyword arguments to pass onto the text object.

    Other Parameters
    ----------------
    **lineprops
        `matplotlib.lines.Line2D` properties that control the appearance of the
        lines. See also `matplotlib.axes.Axes.axhline`.

    """

    def __init__(
        self,
        ax: matplotlib.axes.Axes,
        horizOn: bool = True,
        vertOn: bool = True,
        textOn: bool = True,
        useblit: bool = True,
        textprops: dict | None = None,
        **lineprops: dict,
    ):
        super().__init__(ax)

        if textprops is None:
            textprops = {}

        self.connect_event("motion_notify_event", self.onmove)
        self.connect_event("draw_event", self.clear)

        self.visible = True
        self.horizOn = horizOn
        self.vertOn = vertOn
        self.textOn = textOn
        self.useblit = useblit and self.canvas.supports_blit

        if self.useblit:
            lineprops["animated"] = True
            textprops["animated"] = True

        lcolor = lineprops.pop("color", lineprops.pop("c", "0.35"))
        ls = lineprops.pop("ls", lineprops.pop("linestyle", "--"))
        lw = lineprops.pop("lw", lineprops.pop("linewidth", 0.5))
        lineprops.update({"color": lcolor, "ls": ls, "lw": lw, "visible": False})

        tcolor = textprops.pop("color", textprops.pop("c", lcolor))
        textprops.update(
            {
                "color": tcolor,
                "visible": False,
                "horizontalalignment": "right",
                "verticalalignment": "top",
                "transform": ax.transAxes,
            }
        )

        self.lineh = ax.axhline(ax.get_ybound()[0], **lineprops)
        self.linev = ax.axvline(ax.get_xbound()[0], **lineprops)
        with plt.rc_context({"text.usetex": False}):
            self.label = ax.text(0.95, 0.95, "", **textprops)
        self.background = None
        self.needclear = False

    def clear(self, event):
        """Internal event handler to clear the cursor."""
        if self.ignore(event):
            return
        if self.useblit:
            self.background = self.canvas.copy_from_bbox(self.ax.bbox)
        self.linev.set_visible(False)
        self.lineh.set_visible(False)
        self.label.set_visible(False)

    def onmove(self, event):
        """Internal event handler to draw the cursor when the mouse moves."""
        if self.ignore(event):
            return
        if not self.canvas.widgetlock.available(self):
            return
        if event.inaxes != self.ax:
            self.linev.set_visible(False)
            self.lineh.set_visible(False)
            self.label.set_visible(False)

            if self.needclear:
                self.canvas.draw()
                self.needclear = False
            return
        self.needclear = True
        if not self.visible:
            return
        self.linev.set_xdata((event.xdata, event.xdata))
        self.lineh.set_ydata((event.ydata, event.ydata))
        self.label.set_text(f"({event.xdata:1.3f}, {event.ydata:1.3f})")
        self.linev.set_visible(self.visible and self.vertOn)
        self.lineh.set_visible(self.visible and self.horizOn)
        self.label.set_visible(self.visible and self.textOn)

        self._update()

    def _update(self):
        if self.useblit:
            if self.background is not None:
                self.canvas.restore_region(self.background)
            self.ax.draw_artist(self.linev)
            self.ax.draw_artist(self.lineh)
            self.ax.draw_artist(self.label)
            self.canvas.blit(self.ax.bbox)
        else:
            self.canvas.draw_idle()
        return False


def place_inset(
    parent_axes: matplotlib.axes.Axes,
    width: float | str,
    height: float | str,
    pad: float | tuple[float, float] = 0.1,
    loc: Literal[
        "upper left",
        "upper center",
        "upper right",
        "center left",
        "center",
        "center right",
        "lower left",
        "lower center",
        "lower right",
    ] = "upper right",
    **kwargs: dict,
) -> matplotlib.axes.Axes:
    """Easy placement of inset axes.

    Parameters
    ----------
    parent_axes
        `matplotlib.axes.Axes` to place the inset axes.
    width, height
        Size of the inset axes to create. If `float`, specifies the size in inches, e.g.
        ``1.3``. If `str`, specifies the size in relative units, e.g. ``'40%'`` of
        `parent_axes`.
    pad
        Padding between `parent_axes` and inset in inches.
    loc
        Location to place the inset axes.
    **kwargs
        Keyword arguments are passed onto `matplotlib.axes.Axes.inset_axes`.

    Returns
    -------
    matplotlib.axes.Axes


    """

    return _ez_inset(parent_axes, width, height, pad, loc, **kwargs)


def array_extent(data: xr.DataArray) -> tuple[float, float, float, float]:
    """
    Gets the extent of a :class:`xarray.DataArray` to be used in
    :func:`matplotlib.pyplot.imshow`.

    Parameters
    ----------
    data
        A two-dimensional :class:`xarray.DataArray`.

    Returns
    -------
    x0, x1, y0, y1 : float

    """

    data_coords = tuple(data[dim].values for dim in data.dims)
    data_incs = tuple(coord[1] - coord[0] for coord in data_coords)
    data_lims = tuple((coord[0], coord[-1]) for coord in data_coords)
    y0, x0 = data_lims[0][0] - 0.5 * data_incs[0], data_lims[1][0] - 0.5 * data_incs[1]
    y1, x1 = (
        data_lims[0][-1] + 0.5 * data_incs[0],
        data_lims[1][-1] + 0.5 * data_incs[1],
    )
    return x0, x1, y0, y1


def plot_array(
    arr: xr.DataArray,
    ax: matplotlib.axes.Axes | None = None,
    *,
    colorbar: bool = False,
    colorbar_kw: dict | None = None,
    gamma: float = 1.0,
    norm: mcolors.Normalize | None = None,
    xlim: float | tuple[float, float] | None = None,
    ylim: float | tuple[float, float] | None = None,
    crop: bool = False,
    rad2deg: bool | Iterable[str] = False,
    func: Callable | None = None,
    func_args: dict | None = None,
    **improps: dict,
) -> matplotlib.image.AxesImage:
    """Plots a 2D :class:`xarray.DataArray` using :func:`matplotlib.pyplot.imshow`.

    Parameters
    ----------
    arr
        A two-dimensional :class:`xarray.DataArray` to be plotted.
    ax
        The target :class:`matplotlib.axes.Axes`.
    colorbar
        Whether to plot a colorbar.
    colorbar_kw
        Keyword arguments passed onto :func:`erlab.plotting.colors.nice_colorbar`.
    xlim, ylim
        If given a sequence of length 2, those values are set as the lower and upper
        limits of each axis. If given a single `float`, the limits are set as ``(-lim,
        lim)``.  If `None`, automatically determines the limits from the data.
    rad2deg
        If `True`, converts some known angle coordinates from radians to degrees. If an
        iterable of `str` is given, only the coordinates that correspond to the given
        strings are converted.
    func
        A callable that processes the values prior to display. Its output must have the
        same shape as the input.
    func_args
        Keyword arguments passed onto `func`.
    **improps
        Keyword arguments passed onto :func:`matplotlib.axes.Axes.imshow`.

    Returns
    -------
    matplotlib.image.AxesImage

    """
    if colorbar_kw is None:
        colorbar_kw = {}
    if func_args is None:
        func_args = {}

    if isinstance(arr, np.ndarray):
        arr = xr.DataArray(arr)
    if ax is None:
        ax = plt.gca()
    if xlim is not None and not np.iterable(xlim):
        xlim = (-xlim, xlim)
    if ylim is not None and not np.iterable(ylim):
        ylim = (-ylim, ylim)
    if rad2deg is not False:
        if np.iterable(rad2deg):
            conv_dims = rad2deg
        else:
            conv_dims = [
                d for d in ["phi", "theta", "beta", "alpha", "chi"] if d in arr.dims
            ]
        arr = arr.assign_coords({d: np.rad2deg(arr[d]) for d in conv_dims})

    norm_kw = {}
    if "vmin" in improps:
        norm_kw["vmin"] = improps.pop("vmin")
        if "vmax" in improps:
            norm_kw["vmax"] = improps.pop("vmax")
            colorbar_kw.setdefault("extend", "both")
        else:
            colorbar_kw.setdefault("extend", "min")
    elif "vmax" in improps:
        norm_kw["vmax"] = improps.pop("vmax")
        colorbar_kw.setdefault("extend", "max")

    if norm is None:
        norm = copy.deepcopy(mcolors.PowerNorm(gamma, **norm_kw))

    improps_default = {
        "interpolation": "none",
        "extent": array_extent(arr),
        "aspect": "auto",
        "origin": "lower",
        "rasterized": True,
    }
    for k, v in improps_default.items():
        improps.setdefault(k, v)

    if crop:
        if xlim is not None:
            arr = arr.copy(deep=True).sel({arr.dims[1]: slice(*xlim)})
        if ylim is not None:
            arr = arr.copy(deep=True).sel({arr.dims[0]: slice(*ylim)})
    if func is not None:
        img = ax.imshow(func(arr.values, **func_args), norm=norm, **improps)
    else:
        img = ax.imshow(arr.values, norm=norm, **improps)
    ax.set_xlabel(arr.dims[1])
    ax.set_ylabel(arr.dims[0])
    fancy_labels(ax)
    if xlim is not None:
        ax.set_xlim(*xlim)
    if ylim is not None:
        ax.set_ylim(*ylim)
    if colorbar:
        nice_colorbar(ax=ax, **colorbar_kw)
    return img


def plot_array_2d(
    larr: xr.DataArray,
    carr: xr.DataArray,
    ax: matplotlib.axes.Axes | None = None,
    *,
    xlim: float | tuple[float, float] | None = None,
    ylim: float | tuple[float, float] | None = None,
    cmap: mcolors.Colormap | str = None,
    lnorm: mcolors.Normalize | None = None,
    cnorm: mcolors.Normalize | None = None,
    background: Sequence[float] | None = None,
    colorbar: bool = True,
    cax: matplotlib.axes.Axes | None = None,
    colorbar_kw: dict | None = None,
    imshow_kw: dict | None = None,
    **indexers_kwargs: dict,
):
    if lnorm is None:
        lnorm = plt.Normalize()
    else:
        lnorm = copy.deepcopy(lnorm)
    if cnorm is None:
        cnorm = plt.Normalize()
    else:
        cnorm = copy.deepcopy(cnorm)
    if colorbar_kw is None:
        colorbar_kw = {}
    if imshow_kw is None:
        imshow_kw = {}
    improps_default = {
        "interpolation": "none",
        "aspect": "equal",
        "origin": "lower",
        "rasterized": True,
    }
    for k, v in improps_default.items():
        imshow_kw.setdefault(k, v)
    if ax is None:
        ax = plt.gca()

    larr = larr.qsel(**indexers_kwargs).copy(deep=True)
    carr = carr.qsel(**indexers_kwargs).copy(deep=True)
    sel_kw = {}
    if xlim is not None:
        if not np.iterable(xlim):
            xlim = (-xlim, xlim)
        sel_kw[larr.dims[1]] = slice(*xlim)
    if ylim is not None:
        if not np.iterable(ylim):
            ylim = (-ylim, ylim)
        sel_kw[larr.dims[0]] = slice(*ylim)
    larr = larr.sel(**sel_kw)
    carr = carr.sel(**sel_kw)

    cmap_img, img = gen_2d_colormap(
        larr.values, carr.values, cmap, lnorm=lnorm, cnorm=cnorm, background=background
    )

    if colorbar:
        if cax is None:
            colorbar_kw.setdefault("aspect", 2)
            colorbar_kw.setdefault("anchor", (0, 1))
            colorbar_kw.setdefault("panchor", (0, 1))
            cb = ax.get_figure().colorbar(plt.cm.ScalarMappable(), ax=ax, **colorbar_kw)
            cax = cb.ax
            cax.clear()

        cax.imshow(
            cmap_img.transpose(1, 0, 2),
            extent=(lnorm.vmin, lnorm.vmax, cnorm.vmin, cnorm.vmax),
            origin="lower",
        )

    im = ax.imshow(img, extent=array_extent(larr), **imshow_kw)
    ax.set_xlabel(larr.dims[0])
    ax.set_ylabel(larr.dims[1])

    if colorbar:
        return im, cb
    else:
        return im, None


def gradient_fill(
    x: Sequence[int | float],
    y: Sequence[int | float],
    y0: float | None = None,
    color: str | tuple[float, float, float] | tuple[float, float, float, float] = "C0",
    cmap: str | mcolors.Colormap | None = None,
    transpose: bool = False,
    reverse: bool = False,
    ax: matplotlib.axes.Axes | None = None,
    **kwargs: dict,
) -> matplotlib.image.AxesImage:
    """Applies a gradient fill to a line plot.

    Parameters
    ----------
    x, y
        Data of the plot to fill under.
    y0
        The minimum y value of the gradient. If `None`, defaults to the minimum of `y`.
    color
        A valid matplotlib color to make the gradient from.
    cmap
        If given, ignores `color` and fills with the given colormap.
    transpose
        Transpose the gradient.
    reverse
        Reverse the gradient.
    ax
        The :class:`matplotlib.axes.Axes` to plot in.
    **kwargs
        Keyword arguments passed onto :func:`matplotlib.axes.Axes.imshow`.

    Returns
    -------
    matplotlib.image.AxesImage

    """
    color = kwargs.pop("c", color)
    kwargs.setdefault("norm", InversePowerNorm(0.5))
    kwargs.setdefault("alpha", 0.75)
    if cmap is None:
        cmap = mcolors.LinearSegmentedColormap.from_list(
            "", colors=[(1, 1, 1, 0), mcolors.to_rgba(color)], N=1024
        )
    if isinstance(cmap, str):
        cmap = matplotlib.colormaps[cmap]
    if reverse:
        cmap = cmap.reversed()

    if ax is None:
        ax = plt.gca()
    x, y = np.asarray(x), np.asarray(y)
    valid_inds = (~np.isnan(x)) * (~np.isnan(y))
    x, y = x[valid_inds], y[valid_inds]

    if y0 is None:
        y0 = min(y)
    xn = np.r_[x[0], x, x[-1]]
    yn = np.r_[y0, y, y0]
    patch = matplotlib.patches.PathPatch(
        matplotlib.path.Path(np.array([xn, yn]).T), edgecolor="none", facecolor="none"
    )
    ax.add_patch(patch)

    im = matplotlib.image.AxesImage(
        ax, cmap=cmap, interpolation="bicubic", origin="lower", zorder=0, **kwargs
    )
    im.use_sticky_edges = False
    ax.add_artist(im)
    if transpose:
        im.set_data(np.linspace(0, 1, 1024).reshape(1024, 1).T)
    else:
        im.set_data(np.linspace(0, 1, 1024).reshape(1024, 1))
    # with autoscale_off(ax):
    im.set_extent((min(xn), max(xn), min(yn), max(yn)))
    im.set_clip_path(patch)

    return im


def plot_slices(
    maps: xr.DataArray | Sequence[xr.DataArray],
    figsize: tuple[float, float] | None = None,
    *,
    transpose: bool = False,
    xlim: float | tuple[float, float] | None = None,
    ylim: float | tuple[float, float] | None = None,
    crop: bool = True,
    same_limits: bool = False,
    axis: Literal[
        "on", "off", "equal", "scaled", "tight", "auto", "image", "scaled", "square"
    ] = "auto",
    show_all_labels: bool = False,
    colorbar: Literal["none", "right", "rightspan", "all"] = "none",
    hide_colorbar_ticks: bool = True,
    annotate: bool = True,
    cmap: str | mcolors.Colormap | Iterable[mcolors.Colormap | str] | None = None,
    norm: mcolors.Normalize | Iterable[mcolors.Normalize] | None = None,
    order: Literal["C", "F"] = "C",
    cmap_order: Literal["C", "F"] = "C",
    norm_order: Literal["C", "F"] | None = None,
    gradient: bool = False,
    gradient_kw: dict | None = None,
    subplot_kw: dict | None = None,
    annotate_kw: dict | None = None,
    colorbar_kw: dict | None = None,
    axes: npt.NDArray[matplotlib.axes.Axes] | None = None,
    **values: dict,
) -> tuple[matplotlib.figure.Figure, npt.NDArray[matplotlib.axes.Axes]]:
    """Automated comparison plot of slices.

    Parameters
    ----------
    maps
        Arrays to plot.
    figsize
        Figure size.
    transpose
        Transpose each map before plotting.
    xlim, ylim
        If given a sequence of length 2, those values are set as the lower and upper
        limits of each axis. If given a single `float`, the limits are set as ``(-lim,
        lim)``.  If `None`, automatically determines the limits from the data.
    crop
        If `True`, crops the data to the limits given by `xlim` and `ylim` prior to
        plotting.
    same_limits
        If `True`, all images will have the same vmin and vmax.
    axis
        Passed onto :func:`matplotlib.axes.Axes.axis`. Possible values are:

        ======== ============================================================
        Value    Description
        ======== ============================================================
        'on'     Turn on axis lines and labels.
        'off'    Turn off axis lines and labels.
        'equal'  Set equal scaling (i.e., make circles circular) by
                 changing axis limits. This is the same as ``ax.set_aspect('equal',
                 adjustable='datalim')``. Explicit data limits may not be respected in
                 this case.
        'scaled' Set equal scaling (i.e., make circles circular) by changing dimensions
                 of the plot box. This is the same as ``ax.set_aspect('equal',
                 adjustable='box', anchor='C')``. Additionally, further autoscaling will
                 be disabled.
        'tight'  Set limits just large enough to show all data, then
                 disable further autoscaling.
        'auto'   Automatic scaling (fill plot box with data).
        'image'  'scaled' with axis limits equal to data limits.
        'square' Square plot; similar to 'scaled', but initially forcing
                 ``xmax-xmin == ymax-ymin``.
        ======== ============================================================
    show_all_labels
        If `True`, shows every xlabel and ylabel. If `False`, labels on shared axes are
        minimized. When `False` and the `axes` argument is given, the `order` must be
        specified to correctly hide shared labels.
    colorbar
        Controls colorbar behavior. Possible values are:

        =========== =========================================================
        Value       Description
        =========== =========================================================
        'none'      Do not show colorbars.
        'right'     Creates a colorbar on the right for each row.
        'rightspan' Create a single colorbar that spans all axes.
        'all'       Plot a colorbar for every axes.
        =========== =========================================================
    hide_colorbar_ticks
        If `True`, hides colorbar ticks.
    annotate
        If `False`, turn off automatic annotation.
    cmap
        If supplied a single `str` or :class:`matplotlib.colors.Colormap`, the colormap
        is applied to all axes. Otherwise, a nested sequence with the same shape as the
        resulting axes can be provided to use different colormaps for different axes. If
        the slices are 1D, this argument can be used to supply valid colors as line
        colors for differnet slices.
    norm
        If supplied a single :class:`matplotlib.colors.Normalize`, the norm is applied
        to all axes. Otherwise, a nested sequence with the same shape as the resulting
        axes can be provided to use different norms for different axes.
    order
        Order to display the data. Effectively, this determines if each map is displayed
        along the same row or the same column. 'C' means to flatten in row-major
        (C-style) order, and 'F' means to flatten in column-major (Fortran-style) order.
    cmap_order
        The order to flatten when given a nested sequence for `cmap`,
        Defaults to `order`.
    norm_order
        The order to flatten when given a nested sequence for `norm`,
        Defaults to `cmap_order`.
    gradient
        If `True`, for 1D slices, fills the area under the curve with a gradient. Has no
        effect for 2D slices.
    gradient_kw
        Extra arguments to :func:`gradient_fill`.
    subplot_kw
        Extra arguments to :func:`matplotlib.pyplot.subplots`: refer to the
        `matplotlib` documentation for a list of all possible arguments.
    annotate_kw
        Extra arguments to :func:`erlab.plotting.annotations.label_subplot_properties`.
        Only applied when `annotate` is `True`.
    colorbar_kw
        Extra arguments to :func:`erlab.plotting.colors.proportional_colorbar`.
    axes
        A nested sequence of :class:`matplotlib.axes.Axes`. If supplied, the returned
        :class:`matplotlib.figure.Figure` is inferred from the first axes.
    **values
        Key-value pair of cut location and bin widths. See examples.
        Remaining arguments are passed onto :func:`plot_array`.

    Returns
    -------
    fig : matplotlib.figure.Figure

    axes : array-like of matplotlib.axes.Axes

    Examples
    --------
    ::

        # Two maps: map1, map2
        # Create a figure with a 3 by 2 grid.

        fig, axes = plot_slices([map1, map2], eV=[0, -0.1, -0.2], eV_width=0.05)

    """
    if isinstance(maps, xr.DataArray):
        maps = [maps]

    if transpose:
        maps = [m.T for m in maps]

    if cmap_order is None:
        cmap_order = order

    if norm_order is None:
        norm_order = cmap_order

    if gradient_kw is None:
        gradient_kw = {}

    if subplot_kw is None:
        subplot_kw = {}

    if annotate_kw is None:
        annotate_kw = {}

    if colorbar_kw is None:
        colorbar_kw = {}

    dims = maps[0].dims
    kwargs = {k: v for k, v in values.items() if k not in dims}
    slice_kw = {k: v for k, v in values.items() if k not in kwargs}

    if len(slice_kw) == 0:
        slice_dim = None
        slice_levels = [None]
        slice_width = None

    else:
        slice_dim = next(k for k in slice_kw if not k.endswith("_width"))
        slice_levels = slice_kw[slice_dim]
        slice_width = kwargs.pop(slice_dim + "_width", None)

    plot_dims = [d for d in dims if d != slice_dim]

    if len(plot_dims) not in (1, 2):
        raise ValueError("The data to plot must be 1D or 2D")

    if not np.iterable(slice_levels):
        slice_levels = [slice_levels]

    if xlim is not None and not np.iterable(xlim):
        xlim = (-xlim, xlim)

    if ylim is not None and not np.iterable(ylim):
        ylim = (-ylim, ylim)

    auto_gradient_color = all(k not in gradient_kw for k in ("c", "color"))

    if hide_colorbar_ticks:
        colorbar_kw["ticks"] = []

    for k, v in {"sharex": "col", "sharey": "row", "layout": "constrained"}.items():
        subplot_kw.setdefault(k, v)

    if order == "F":
        nrow, ncol = len(slice_levels), len(maps)

    elif order == "C":
        nrow, ncol = len(maps), len(slice_levels)

    cmap_name = cmap
    cmap_norm = norm
    if axes is None:
        fig, axes = plt.subplots(nrow, ncol, figsize=figsize, **subplot_kw)

    else:
        fig = axes.flat[0].get_figure()

    if nrow == 1:
        axes = axes[:, np.newaxis].reshape(1, -1)

    if ncol == 1:
        axes = axes[:, np.newaxis].reshape(-1, 1)

    qsel_kw = {}

    if crop:
        if len(plot_dims) == 1:
            if transpose and (ylim is not None):
                qsel_kw[plot_dims[0]] = slice(*ylim)

            elif xlim is not None:
                qsel_kw[plot_dims[0]] = slice(*xlim)

        elif len(plot_dims) == 2:
            if xlim is not None:
                qsel_kw[plot_dims[1]] = slice(*xlim)

            if ylim is not None:
                qsel_kw[plot_dims[0]] = slice(*ylim)

    if slice_width is not None:
        qsel_kw[slice_dim + "_width"] = slice_width

    for i in range(len(slice_levels)):
        if slice_dim is not None:
            qsel_kw[slice_dim] = slice_levels[i]

        for j in range(len(maps)):
            dat_sel = maps[j].copy(deep=True).qsel(**qsel_kw)

            if order == "F":
                ax = axes[i, j]
            elif order == "C":
                ax = axes[j, i]

            if np.iterable(cmap_name) and not isinstance(cmap_name, str):
                if cmap_order == "F":
                    if isinstance(cmap_name[i], str):
                        cmap = cmap_name[i]
                    else:
                        cmap = cmap_name[i][j]
                elif cmap_order == "C":
                    if isinstance(cmap_name[j], str):
                        cmap = cmap_name[j]
                    else:
                        cmap = cmap_name[j][i]
            else:
                cmap = cmap_name

            if len(plot_dims) == 1:
                if cmap is not None:
                    kwargs["color"] = cmap
                    if auto_gradient_color:
                        gradient_kw["color"] = cmap

                if transpose:
                    ax.plot(dat_sel.values, dat_sel[plot_dims[0]], **kwargs)
                    ax.set_xlabel(dat_sel.name)
                    ax.set_ylabel(plot_dims[0])

                    if gradient:
                        gradient_fill(
                            dat_sel.values,
                            dat_sel[plot_dims[0]],
                            ax=ax,
                            transpose=True,
                            **gradient_kw,
                        )

                else:
                    dat_sel.plot(ax=ax, **kwargs)
                    ax.set_title("")

                    if gradient:
                        gradient_fill(
                            dat_sel[plot_dims[0]], dat_sel.values, ax=ax, **gradient_kw
                        )

            elif len(plot_dims) == 2:
                if np.iterable(cmap_norm):
                    if norm_order == "F":
                        try:
                            norm = cmap_norm[i][j]
                        except TypeError:
                            norm = cmap_norm[i]

                    elif norm_order == "C":
                        try:
                            norm = cmap_norm[j][i]
                        except TypeError:
                            norm = cmap_norm[j]
                else:
                    norm = copy.deepcopy(cmap_norm)
                plot_array(dat_sel, ax=ax, norm=norm, cmap=cmap, **kwargs)

    if same_limits and len(plot_dims) == 2:
        vmn, vmx = [], []
        for ax in axes.flat:
            vmn.append(ax.get_images()[0].norm.vmin)
            vmx.append(ax.get_images()[0].norm.vmax)
        vmn, vmx = min(vmn), max(vmx)
        for ax in axes.flat:
            ax.get_images()[0].norm.vmin = vmn
            ax.get_images()[0].norm.vmax = vmx

    for ax in axes.flatten():
        if not show_all_labels:
            if ax not in axes[:, 0]:
                ax.set_ylabel("")
            if ax not in axes[-1, :]:
                ax.set_xlabel("")
        if not gradient:
            ax.axis(axis)
        if xlim is not None:
            ax.set_xlim(*xlim)
        if ylim is not None:
            ax.set_ylim(*ylim)
        if colorbar not in ["none", "rightspan"]:
            if colorbar == "all" or ax in axes[:, -1]:
                nice_colorbar(mappable=ax.images[0], ax=ax, **colorbar_kw)
    if colorbar == "rightspan":
        nice_colorbar(ax=axes, **colorbar_kw)
    if annotate and slice_dim is not None:
        if slice_dim == "eV":
            slice_dim = "Eb"
        label_subplot_properties(
            axes,
            values={slice_dim: slice_levels * len(maps)},
            order=order,
            **annotate_kw,
        )
    return fig, axes


def fermiline(
    ax: matplotlib.axes.Axes | None = None,
    value: float = 0.0,
    orientation: Literal["h", "v"] = "h",
    **kwargs: dict,
) -> matplotlib.lines.Line2D:
    """Plots a constant energy line to denote the Fermi level.

    Parameters
    ----------
    ax
        The `matplotlib.axes.Axes` to annotate.
    value
        The coordinate of the line. Defaults to 0, assuming binding energy.
    orientation
        If 'h', a horizontal line is plotted. If 'v', a vertical line is plotted.
    **kwargs
        Keyword arguments passed onto `matplotlib.lines.Line2D`.

    Returns
    -------
    matplotlib.lines.Line2D

    """
    if ax is None:
        ax = plt.gca()
    if np.iterable(ax):
        return [fermiline(a, value, orientation, **kwargs) for a in np.asarray(ax).flat]

    c = kwargs.pop("color", kwargs.pop("c", axes_textcolor(ax)))
    lw = kwargs.pop("lw", kwargs.pop("linewidth", 0.25))
    ls = kwargs.pop("ls", kwargs.pop("linestyle", "-"))
    match orientation:
        case "h":
            return ax.axhline(value, ls=ls, lw=lw, c=c, **kwargs)
        case "v":
            return ax.axvline(value, ls=ls, lw=lw, c=c, **kwargs)
        case _:
            raise ValueError("`orientation` must be either 'v' or 'h'")



"""
Polygon mask generation code adapted from the `CGAL C++ library
<https://doc.cgal.org/5.3.2/Polygon/index.html>`_.

"""

from __future__ import annotations

from typing import Annotated, Literal

import enum

import numpy as np
import numpy.typing as npt
import numba


class Comparison(enum.Enum):
    SMALLER = -1
    EQUAL = 0
    LARGER = 1


class Side(enum.Enum):
    ON_UNBOUNDED_SIDE = -1
    ON_BOUNDARY = 0
    ON_BOUNDED_SIDE = 1


@numba.njit(nogil=True, cache=True)
def _comp_x(p0, p1):
    if np.float32(p0[0]) < np.float32(p1[0]):
        return Comparison.SMALLER
    elif np.float32(p0[0]) == np.float32(p1[0]):
        return Comparison.EQUAL
    else:
        return Comparison.LARGER


@numba.njit(nogil=True, cache=True)
def _comp_y(p0, p1):
    if np.float32(p0[1]) < np.float32(p1[1]):
        return Comparison.SMALLER
    elif np.float32(p0[1]) == np.float32(p1[1]):
        return Comparison.EQUAL
    else:
        return Comparison.LARGER


@numba.njit(nogil=True, cache=True)
def _get_argmin_all(arr):
    return np.nonzero(arr == arr.min())[0]


@numba.njit(nogil=True, cache=True)
def _get_argmax_all(arr):
    return np.nonzero(arr == arr.max())[0]


@numba.njit(nogil=True, cache=True)
def left_vertex(points: Annotated[npt.NDArray[np.float64], Literal[1, 2]]) -> int:
    """Returns the index of the leftmost point of a polygon.

    In case of a tie, the point with the smallest y-coordinate is taken.

    Parameters
    ----------
    points
        Input array of polygon vertices.

    Returns
    -------
    int

    """
    ind = _get_argmin_all(points[:, 0].astype(np.float32))
    return ind[np.argmin(points[ind][:, 1])]


@numba.njit(nogil=True, cache=True)
def right_vertex(points: Annotated[npt.NDArray[np.float64], Literal[1, 2]]):
    """Returns the index of the rightmost point of a polygon.

    In case of a tie, the point with the largest y-coordinate is taken.

    Parameters
    ----------
    points
        Input array of polygon vertices.

    Returns
    -------
    int

    """
    ind = _get_argmax_all(points[:, 0].astype(np.float32))
    return ind[np.argmax(points[ind][:, 1])]


@numba.njit(nogil=True, cache=True)
def polygon_orientation(points):
    ind = left_vertex(points)
    return _orientation(points[ind - 1], points[ind], points[ind + 1])


@numba.njit(nogil=True, cache=True)
def _orientation(p0, p1, p2):
    # 1: left turn
    # -1: right turn
    # 0: colinear
    p10, p20 = p1 - p0, p2 - p0
    return np.sign(p10[0] * p20[1] - p20[0] * p10[1])


@numba.njit(nogil=True, cache=True)
def which_side_in_slab(point, low, high, points):
    point = np.asarray(point)
    low_x_comp_res = _comp_x(point, points[low])
    high_x_comp_res = _comp_x(point, points[high])

    if low_x_comp_res == Comparison.SMALLER:
        if high_x_comp_res == Comparison.SMALLER:
            return -1
    else:
        match high_x_comp_res:
            case Comparison.LARGER:
                return 1
            case Comparison.SMALLER:
                pass
            case Comparison.EQUAL:
                return int(low_x_comp_res != Comparison.EQUAL)

    return _orientation(points[low], point, points[high])


@numba.njit(nogil=True, cache=True)
def bounded_side_bool(
    points: npt.NDArray[np.float64], point: tuple[float, float], boundary=True
):
    match bounded_side(points, point):
        case Side.ON_UNBOUNDED_SIDE:
            return False
        case Side.ON_BOUNDED_SIDE:
            return True
        case Side.ON_BOUNDARY:
            return boundary


@numba.njit(nogil=True, cache=True)
def bounded_side(points: npt.NDArray[np.float64], point: tuple[float, float]):
    """Computes if a point lies inside a polygon.

    The polygon is defined by the sequence of points [first,last). Being inside is
    defined by the odd-even rule. If the point is on a polygon edge, a special value is
    returned. A simple polygon divides the plane in an unbounded and a bounded region.
    According to the definition points in the bounded region are inside the polygon.

    Parameters
    ----------
    points
        (N, 2) input array of polygon vertices.
    point
        2-tuple specifying point of interest

    Note
    ----
    We shoot a horizontal ray from the point to the right and count the number of
    intersections with polygon segments. If the number of intersections is odd, the
    point is inside. We don't count intersections with horizontal segments. With
    non-horizontal segments, the top vertex is considered to be part of the segment, but
    the bottom vertex is not. (Segments are half-closed).

    """

    last = len(points) - 1
    if last < 2:
        return Side.ON_UNBOUNDED_SIDE

    is_inside = False

    cur_y_comp_res = _comp_y(points[0], point)
    for i in range(len(points) - 1):
        next_y_comp_res = _comp_y(points[i + 1], point)

        match cur_y_comp_res:
            case Comparison.SMALLER:
                match next_y_comp_res:
                    case Comparison.SMALLER:
                        pass

                    case Comparison.EQUAL:
                        match _comp_x(point, points[i + 1]):
                            case Comparison.SMALLER:
                                is_inside = not is_inside
                            case Comparison.EQUAL:
                                return Side.ON_BOUNDARY
                            case Comparison.LARGER:
                                pass

                    case Comparison.LARGER:
                        match which_side_in_slab(point, i, i + 1, points):
                            case -1:
                                is_inside = not is_inside
                            case 0:
                                return Side.ON_BOUNDARY

            case Comparison.EQUAL:
                match next_y_comp_res:
                    case Comparison.SMALLER:
                        match _comp_x(point, points[i]):
                            case Comparison.SMALLER:
                                is_inside = not is_inside
                            case Comparison.EQUAL:
                                return Side.ON_BOUNDARY
                            case Comparison.LARGER:
                                pass

                    case Comparison.EQUAL:
                        match _comp_x(point, points[i]):
                            case Comparison.SMALLER:
                                if _comp_x(point, points[i + 1]) != Comparison.SMALLER:
                                    return Side.ON_BOUNDARY
                            case Comparison.EQUAL:
                                return Side.ON_BOUNDARY
                            case Comparison.LARGER:
                                if _comp_x(point, points[i + 1]) != Comparison.LARGER:
                                    return Side.ON_BOUNDARY

                    case Comparison.LARGER:
                        if _comp_x(point, points[i]) == Comparison.EQUAL:
                            return Side.ON_BOUNDARY

            case Comparison.LARGER:
                match next_y_comp_res:
                    case Comparison.SMALLER:
                        match which_side_in_slab(point, i + 1, i, points):
                            case -1:
                                is_inside = not is_inside
                            case 0:
                                return Side.ON_BOUNDARY
                    case Comparison.EQUAL:
                        if _comp_x(point, points[i + 1]) == Comparison.EQUAL:
                            return Side.ON_BOUNDARY
                    case Comparison.LARGER:
                        pass

        cur_y_comp_res = next_y_comp_res

    if is_inside:
        return Side.ON_BOUNDED_SIDE
    else:
        return Side.ON_UNBOUNDED_SIDE



"""Utilities related to manipulating colors.

Colormaps
---------

In addition to the default `matplotlib
<https://matplotlib.org/stable/tutorials/colors/colormaps.html>`_ colormaps, `cmasher
<https://cmasher.readthedocs.io>`_, `cmocean <https://matplotlib.org/cmocean/>`_, and
`colorcet <https://colorcet.holoviz.org>`_ packages can be installed to extend the
available colormaps. If these packages are installed, they will be automatically
imported upon importing `erlab.plotting`.

Colormap Normalization
----------------------

.. plot:: norms.py example_1
   :width: 65 %

   Demonstration of `InversePowerNorm`.

.. plot:: norms.py example_2
   :width: 65 %

   Demonstration of `CenteredPowerNorm` and `CenteredInversePowerNorm`.

"""

__all__ = [
    "InversePowerNorm",
    "TwoSlopePowerNorm",
    "CenteredPowerNorm",
    "TwoSlopeInversePowerNorm",
    "CenteredInversePowerNorm",
    "get_mappable",
    "unify_clim",
    "proportional_colorbar",
    "nice_colorbar",
    "flatten_transparency",
    "gen_2d_colormap",
    "color_distance",
    "close_to_white",
    "prominent_color",
    "image_is_light",
    "axes_textcolor",
]

from collections.abc import Iterable, Sequence
from numbers import Number
from typing import Literal

import matplotlib
import matplotlib.axes
import matplotlib.colors as mcolors
import matplotlib.pyplot as plt
import matplotlib.transforms as mtransforms
import numpy as np


class InversePowerNorm(mcolors.Normalize):
    r"""
    Linearly map a given value to the 0-1 range and then apply an inverse power-law
    normalization over that range.

    For values :math:`x`, `matplotlib.colors.PowerNorm` calculates
    :math:`x^\gamma`, whereas `InversePowerNorm` calculates :math:`1-x^{1/\gamma}`.
    This provides higher contrast for values closer to ``vmin``.

    Parameters
    ----------
    gamma
        Power law normalization parameter. If equal to 1, the colormap is linear.
    vmin, vmax
        If ``vmin`` and/or ``vmax`` is not given, they are initialized from the
        minimum and maximum value, respectively, of the first input
        processed; i.e., ``__call__(A)`` calls ``autoscale_None(A)``
    clip
        If ``True`` values falling outside the range ``[vmin, vmax]``,
        are mapped to 0 or 1, whichever is closer, and masked values are
        set to 1.  If ``False`` masked values remain masked.

        Clipping silently defeats the purpose of setting the over, under,
        and masked colors in a colormap, so it is likely to lead to
        surprises; therefore the default is ``clip=False``.

    """

    def __init__(
        self,
        gamma: float,
        vmin: float | None = None,
        vmax: float | None = None,
        clip: bool = False,
    ):
        super().__init__(vmin, vmax, clip)
        self.gamma = gamma

    def __call__(self, value, clip=None):
        if clip is None:
            clip = self.clip

        result, is_scalar = self.process_value(value)

        self.autoscale_None(result)
        gamma = self.gamma
        vmin, vmax = self.vmin, self.vmax
        if vmin > vmax:
            raise ValueError("minvalue must be less than or equal to maxvalue")
        elif vmin == vmax:
            result.fill(0)
        else:
            if clip:
                mask = np.ma.getmask(result)
                result = np.ma.array(
                    np.clip(result.filled(vmax), vmin, vmax), mask=mask
                )
            resdat = result.data

            resdat *= -1
            resdat += vmax
            resdat /= vmax - vmin
            resdat[resdat < 0] = 0
            np.power(resdat, 1.0 / gamma, resdat)
            resdat *= -1
            resdat += 1

            result = np.ma.array(resdat, mask=result.mask, copy=False)
        if is_scalar:
            result = result[0]
        return result

    def inverse(self, value):
        if not self.scaled():
            raise ValueError("Not invertible until scaled")
        gamma = self.gamma
        vmin, vmax = self.vmin, self.vmax

        if np.iterable(value):
            val = np.ma.asarray(value)
            return np.ma.power(1 - val, gamma) * (vmin - vmax) + vmax
        else:
            return pow(1 - value, gamma) * (vmin - vmax) + vmax


def _diverging_powernorm(result, gamma, vmin, vmax, vcenter):
    resdat = result.data
    resdat_ = resdat.copy()

    resdat_l = resdat[resdat_ < vcenter]
    resdat_l -= vcenter
    resdat_l[resdat_l >= 0] = 0
    np.power(-resdat_l, gamma, resdat_l)
    resdat_l /= (vcenter - vmin) ** gamma
    resdat_l *= -0.5
    resdat_l += 0.5
    resdat[resdat_ < vcenter] = resdat_l

    resdat_u = resdat[resdat_ >= vcenter]
    resdat_u -= vcenter
    resdat_u[resdat_u < 0] = 0
    np.power(resdat_u, gamma, resdat_u)
    resdat_u /= (vmax - vcenter) ** gamma
    resdat_u *= 0.5
    resdat_u += 0.5
    resdat[resdat_ >= vcenter] = resdat_u

    result = np.ma.array(resdat, mask=result.mask, copy=False)
    return result


def _diverging_powernorm_inv(value, gamma, vmin, vmax, vcenter):
    if np.iterable(value):
        val = np.ma.asarray(value)
        val_ = val.copy()
        val_l = val_[val < 0.5]
        val_u = val_[val >= 0.5]
        val_[val < 0.5] = (
            np.ma.power(1 - 2 * val_l, 1.0 / gamma) * (vmin - vcenter) + vcenter
        )
        val_[val >= 0.5] = (
            np.ma.power(2 * val_u - 1, 1.0 / gamma) * (vmax - vcenter) + vcenter
        )
        return np.ma.asarray(val_)
    elif value < 0.5:
        return pow(1 - 2 * value, 1.0 / gamma) * (vmin - vcenter) + vcenter
    else:
        return pow(2 * value - 1, 1.0 / gamma) * (vmax - vcenter) + vcenter


def _diverging_inversepowernorm(result, gamma, vmin, vmax, vcenter):
    resdat = result.data
    resdat_ = resdat.copy()

    resdat_l = resdat[resdat_ < vcenter]
    resdat_l *= -1
    resdat_l += vmin
    resdat_l /= vmin - vcenter
    resdat_l[resdat_l < 0] = 0
    np.power(resdat_l, 1.0 / gamma, resdat_l)
    resdat_l *= 0.5
    resdat[resdat_ < vcenter] = resdat_l

    resdat_u = resdat[resdat_ >= vcenter]
    resdat_u *= -1
    resdat_u += vmax
    resdat_u /= vmax - vcenter
    resdat_u[resdat_u < 0] = 0
    np.power(resdat_u, 1.0 / gamma, resdat_u)
    resdat_u *= -0.5
    resdat_u += 1
    resdat[resdat_ >= vcenter] = resdat_u

    result = np.ma.array(resdat, mask=result.mask, copy=False)
    return result


def _diverging_inversepowernorm_inv(value, gamma, vmin, vmax, vcenter):
    if np.iterable(value):
        val = np.ma.asarray(value)
        val_ = val.copy()
        val_l = val_[val < 0.5]
        val_u = val_[val >= 0.5]
        val_[val < 0.5] = np.ma.power(2 * val_l, gamma) * (vcenter - vmin) + vmin
        val_[val >= 0.5] = np.ma.power(2 - 2 * val_u, gamma) * (vcenter - vmax) + vmax
        return np.ma.asarray(val_)
    elif value < 0.5:
        return pow(2 * value, gamma) * (vcenter - vcenter) + vmin
    else:
        return pow(2 - 2 * value, gamma) * (vcenter - vmax) + vmax


class TwoSlopePowerNorm(mcolors.TwoSlopeNorm):
    r"""Power-law normalization of data with a set center.

    Useful when mapping data with an unequal rates of change around a
    conceptual center, e.g., data that range from -2 to 4, with 0 as
    the midpoint.

    Parameters
    ----------
    gamma
        Power law exponent.
    vcenter
        The data value that defines ``0.5`` in the normalization.
        Defaults to ``0``.
    vmin
        The data value that defines ``0.0`` in the normalization.
        Defaults to the min value of the dataset.
    vmax
        The data value that defines ``1.0`` in the normalization.
        Defaults to the max value of the dataset.

    """

    def __init__(
        self,
        gamma: float,
        vcenter: float = 0.0,
        vmin: float | None = None,
        vmax: float | None = None,
    ):
        super().__init__(vcenter=vcenter, vmin=vmin, vmax=vmax)
        self.gamma = gamma
        self._func = _diverging_powernorm
        self._func_i = _diverging_powernorm_inv

    def __call__(self, value, clip=None):
        """
        Map value to the interval [0, 1]. The clip argument is unused.
        """
        result, is_scalar = self.process_value(value)

        self.autoscale_None(result)
        gamma = self.gamma
        vmin, vcenter, vmax = self.vmin, self.vcenter, self.vmax
        if not vmin <= vcenter <= vmax:
            raise ValueError("vmin, vcenter, vmax must increase monotonically")
        if vmin == vmax:
            result.fill(0)
        else:
            result = self._func(result, gamma, vmin, vmax, vcenter)
        if is_scalar:
            result = np.atleast_1d(result)[0]
        return result

    def inverse(self, value):
        if not self.scaled():
            raise ValueError("Not invertible until scaled")
        gamma = self.gamma
        (vmin,), _ = self.process_value(self.vmin)
        (vmax,), _ = self.process_value(self.vmax)
        (vcenter,), _ = self.process_value(self.vcenter)
        return self._func_i(value, gamma, vmin, vmax, vcenter)


class CenteredPowerNorm(mcolors.CenteredNorm):
    r"""Power-law normalization of symmetrical data around a center.

    Unlike `TwoSlopePowerNorm`, `CenteredPowerNorm` applies an equal rate of
    change around the center.

    Useful when mapping symmetrical data around a conceptual center e.g., data that
    range from -2 to 4, with 0 as the midpoint, and with equal rates of change
    around that midpoint.

    Parameters
    ----------
    gamma
        Power law exponent.
    vcenter
        The data value that defines ``0.5`` in the normalization. Defaults to ``0``.
    halfrange
        The range of data values that defines a range of ``0.5`` in the
        normalization, so that `vcenter` - `halfrange` is ``0.0`` and `vcenter` +
        `halfrange` is ``1.0`` in the normalization. Defaults to the largest
        absolute difference to `vcenter` for the values in the dataset.
    clip
        If ``True`` values falling outside the range ``[vmin, vmax]``,
        are mapped to 0 or 1, whichever is closer, and masked values are
        set to 1.  If ``False`` masked values remain masked.

        Clipping silently defeats the purpose of setting the over, under,
        and masked colors in a colormap, so it is likely to lead to
        surprises; therefore the default is ``clip=False``.

    """

    def __init__(
        self,
        gamma: float,
        vcenter: float = 0,
        halfrange: float | None = None,
        clip: bool = False,
    ):
        super().__init__(vcenter=vcenter, halfrange=halfrange, clip=clip)
        self.gamma = gamma
        self._func = _diverging_powernorm
        self._func_i = _diverging_powernorm_inv

    def __call__(self, value, clip=None):
        """
        Map value to the interval [0, 1].
        """
        if clip is None:
            clip = self.clip

        result, is_scalar = self.process_value(value)

        self.autoscale_None(result)
        gamma = self.gamma
        vmin, vcenter, vmax = self.vmin, self.vcenter, self.vmax
        if not vmin <= vcenter <= vmax:
            raise ValueError("vmin, vcenter, vmax must increase monotonically")
        if vmin == vmax:
            result.fill(0)
        else:
            if clip:
                mask = np.ma.getmask(result)
                result = np.ma.array(
                    np.clip(result.filled(vmax), vmin, vmax), mask=mask
                )
            result = self._func(result, gamma, vmin, vmax, vcenter)
        if is_scalar:
            result = np.atleast_1d(result)[0]
        return result

    def inverse(self, value):
        if not self.scaled():
            raise ValueError("Not invertible until scaled")
        gamma = self.gamma
        (vmin,), _ = self.process_value(self.vmin)
        (vmax,), _ = self.process_value(self.vmax)
        (vcenter,), _ = self.process_value(self.vcenter)
        return self._func_i(value, gamma, vmin, vmax, vcenter)


class TwoSlopeInversePowerNorm(TwoSlopePowerNorm):
    r"""Inverse power-law normalization of data with a set center.

    Useful when mapping data with an unequal rates of change around a
    conceptual center, e.g., data that range from -2 to 4, with 0 as
    the midpoint.

    Parameters
    ----------
    gamma
        Power law exponent.
    vcenter
        The data value that defines ``0.5`` in the normalization.
        Defaults to ``0``.
    vmin
        The data value that defines ``0.0`` in the normalization.
        Defaults to the min value of the dataset.
    vmax
        The data value that defines ``1.0`` in the normalization.
        Defaults to the max value of the dataset.

    """

    def __init__(
        self,
        gamma: float,
        vcenter: float = 0.0,
        vmin: float | None = None,
        vmax: float | None = None,
    ):
        super().__init__(gamma, vcenter, vmin, vmax)
        self._func = _diverging_inversepowernorm
        self._func_i = _diverging_inversepowernorm_inv


class CenteredInversePowerNorm(CenteredPowerNorm):
    r"""Inverse power-law normalization of symmetrical data around a center.

    Unlike `TwoSlopeInversePowerNorm`, `CenteredInversePowerNorm` applies an
    equal rate of change around the center.

    Useful when mapping symmetrical data around a conceptual center e.g., data that
    range from -2 to 4, with 0 as the midpoint, and with equal rates of change
    around that midpoint.

    Parameters
    ----------
    gamma
        Power law exponent.
    vcenter
        The data value that defines ``0.5`` in the normalization. Defaults to ``0``.
    halfrange
        The range of data values that defines a range of ``0.5`` in the
        normalization, so that `vcenter` - `halfrange` is ``0.0`` and `vcenter` +
        `halfrange` is ``1.0`` in the normalization. Defaults to the largest
        absolute difference to `vcenter` for the values in the dataset.
    clip
        If ``True`` values falling outside the range ``[vmin, vmax]``,
        are mapped to 0 or 1, whichever is closer, and masked values are
        set to 1.  If ``False`` masked values remain masked.

        Clipping silently defeats the purpose of setting the over, under,
        and masked colors in a colormap, so it is likely to lead to
        surprises; therefore the default is ``clip=False``.

    """

    def __init__(
        self,
        gamma: float,
        vcenter: float = 0,
        halfrange: float | None = None,
        clip: bool = False,
    ):
        super().__init__(gamma, vcenter, halfrange, clip)
        self._func = _diverging_inversepowernorm
        self._func_i = _diverging_inversepowernorm_inv


def get_mappable(
    ax: matplotlib.axes.Axes, image_only: bool = False, silent: bool = False
) -> matplotlib.cm.ScalarMappable | None:
    """Gets the `matplotlib.cm.ScalarMappable` from a given `matplotlib.axes.Axes`.

    Parameters
    ----------
    ax
        Parent axes.
    image_only
        Only consider images as a valid mappable, by default `False`.
    silent
        If `False`, raises a `RuntimeError`. If `True`, silently returns `None`.

    Returns
    -------
    matplotlib.cm.ScalarMappable or None

    """
    try:
        if not image_only:
            mappable = ax.collections[-1]
        else:
            raise AttributeError
    except (IndexError, AttributeError):
        try:
            mappable = ax.get_images()[-1]
        except (IndexError, AttributeError):
            mappable = None
    if not silent and mappable is None:
        raise RuntimeError(
            "No mappable was found to use for colorbar "
            "creation. First define a mappable such as "
            "an image (with imshow) or a contour set ("
            "with contourf)."
        )
    return mappable


def unify_clim(axes, target=None, image_only: bool = False):
    if target is None:
        vmn, vmx = [], []
        for ax in axes.flat:
            mappable = get_mappable(ax, image_only=image_only)
            vmn.append(mappable.norm.vmin)
            vmx.append(mappable.norm.vmax)
        vmn, vmx = min(vmn), max(vmx)
    else:
        mappable = get_mappable(target, image_only=image_only)
        vmn, vmx = mappable.norm.vmin, mappable.norm.vmax
    for ax in axes.flat:
        mappable = get_mappable(ax, image_only=image_only)
        mappable.norm.vmin, mappable.norm.vmax = vmn, vmx


def proportional_colorbar(
    mappable: matplotlib.cm.ScalarMappable | None = None,
    cax: matplotlib.axes.Axes | None = None,
    ax: matplotlib.axes.Axes | Iterable[matplotlib.axes.Axes] | None = None,
    **kwargs: dict,
) -> matplotlib.colorbar.Colorbar:
    """
    Replaces the current colorbar or creates a new colorbar with proportional spacing.

    The default behavior of colorbars in `matplotlib` does not support colors
    proportional to data in different norms. This function circumvents this behavior.

    Parameters
    ----------
    mappable
        The `matplotlib.cm.ScalarMappable` described by this colorbar.
    cax
        Axes into which the colorbar will be drawn.
    ax
        One or more parent axes from which space for a new colorbar axes
        will be stolen, if `cax` is `None`.  This has no effect if `cax`
        is set. If `mappable` is `None` and `ax` is given with more than
        one Axes, the function will try to infer the mappable from the
        first one.
    **kwargs
        Extra arguments to `matplotlib.pyplot.colorbar`: refer to the `matplotlib`
        documentation for a list of all possible arguments.

    Returns
    -------
    cbar : matplotlib.colorbar.Colorbar
        The created colorbar.

    Examples
    --------
    ::

        import numpy as np
        import matplotlib.pyplot as plt
        import matplotlib.colors as colors

        # Create example data and plot
        X, Y = np.mgrid[0 : 3 : complex(0, 100), 0 : 2 : complex(0, 100)]
        pcm = plt.pcolormesh(
            X,
            Y,
            (1 + np.sin(Y * 10.0)) * X**2,
            norm=mcolors.PowerNorm(gamma=0.5),
            cmap="Blues_r",
            shading="auto",
        )

        # Plot evenly spaced colorbar
        proportional_colorbar()

    """
    fontsize = kwargs.pop("fontsize", None)

    if ax is None:
        if cax is None:
            ax = plt.gca()
            if mappable is None:
                mappable = get_mappable(ax)
    elif isinstance(ax, np.ndarray):
        i = 0
        while mappable is None and i < len(ax.flat):
            mappable = get_mappable(ax.flatten()[i], silent=(i != (len(ax.flat) - 1)))
            i += 1
    elif mappable is None:
        mappable = get_mappable(ax)

    if mappable.colorbar is None:
        plt.colorbar(mappable=mappable, cax=cax, ax=ax, **kwargs)
    ticks = mappable.colorbar.get_ticks()
    if cax is None:
        mappable.colorbar.remove()
    kwargs.setdefault("ticks", ticks)
    kwargs.setdefault("cmap", mappable.cmap)
    kwargs.setdefault("norm", mappable.norm)
    kwargs.setdefault("pad", 0.05)
    kwargs.setdefault("fraction", 0.05)
    kwargs.setdefault("aspect", 25)

    cbar = plt.colorbar(
        mappable=mappable,
        cax=cax,
        ax=ax,
        spacing="proportional",
        boundaries=kwargs["norm"].inverse(np.linspace(0, 1, kwargs["cmap"].N)),
        **kwargs,
    )
    if fontsize is not None:
        cbar.ax.tick_params(labelsize=fontsize)
    return cbar


def _size_to_bounds(ax, width, height, loc):
    fig = ax.get_figure()
    sizes = [width, height]

    ax_sizes = (
        ax.get_window_extent().transformed(fig.dpi_scale_trans.inverted()).bounds[2:]
    )
    sizes = [
        float(size[:-1]) / 100 if isinstance(size, str) else size / ax_sizes[i]
        for i, size in enumerate(sizes)
    ]

    origin = [1 - sizes[0], 1 - sizes[1]]
    if "center" in loc:
        origin[0] /= 2
        origin[1] /= 2
        if "upper" in loc or "lower" in loc:
            origin[1] *= 2
        elif "left" in loc or "right" in loc:
            origin[0] *= 2
    if "left" in loc:
        origin[0] = 0
    if "lower" in loc:
        origin[1] = 0
    return origin + sizes


def _refresh_pads(ax, cax, pads, loc):
    ref = _size_to_bounds(ax, 0, 0, loc)[:2]
    bbox = ax.get_window_extent().transformed(ax.figure.dpi_scale_trans.inverted())
    x0, y0 = (
        bbox.x0 + ref[0] * (bbox.x1 - bbox.x0),
        bbox.y1 + ref[1] * (bbox.y1 - bbox.y0),
    )
    bbox = cax.get_window_extent().transformed(ax.figure.dpi_scale_trans.inverted())
    x1, y1 = (
        bbox.x0 + ref[0] * (bbox.x1 - bbox.x0),
        bbox.y1 + ref[1] * (bbox.y1 - bbox.y0),
    )

    pads[0] += x1 - x0
    pads[1] += y1 - y0
    return pads


def _get_pad(pad, loc):
    pad_num = False
    if isinstance(pad, Number):
        pad_num = True
        pad = [pad, pad]
    pads = [-pad[0], -pad[1]]
    if "center" in loc:
        pads[0] *= -1
        pads[1] *= -1
        if "upper" in loc or "lower" in loc:
            if pad_num:
                pads[0] = 0
            pads[1] *= -1
        elif "left" in loc or "right" in loc:
            if pad_num:
                pads[1] = 0
            pads[0] *= -1
    if "left" in loc:
        pads[0] *= -1
    if "lower" in loc:
        pads[1] *= -1
    return pads


def _ez_inset(
    parent_axes: matplotlib.axes.Axes,
    width: float | str,
    height: float | str,
    pad: float | tuple[float, float] = 0.1,
    loc: Literal[
        "upper left",
        "upper center",
        "upper right",
        "center left",
        "center",
        "center right",
        "lower left",
        "lower center",
        "lower right",
    ] = "upper right",
    **kwargs: dict,
) -> matplotlib.axes.Axes:
    fig = parent_axes.get_figure()
    locator = InsetAxesLocator(parent_axes, width, height, pad, loc)
    ax_ = fig.add_axes(locator(parent_axes, None).bounds, **kwargs)
    ax_.set_axes_locator(locator)
    return ax_


class InsetAxesLocator:
    def __init__(self, ax, width, height, pad, loc):
        self._ax = ax
        self._transAxes = ax.transAxes
        self._width = width
        self._height = height
        self._loc = loc
        self.set_pad(pad)

    def __call__(self, ax, renderer):
        return mtransforms.TransformedBbox(
            mtransforms.Bbox.from_bounds(*self._size_to_bounds(ax)),
            self._transAxes
            + mtransforms.ScaledTranslation(*self.pads, ax.figure.dpi_scale_trans)
            - ax.figure.transSubfigure,
        )

    def set_pad(self, pad):
        pad_num = False
        if isinstance(pad, Number):
            pad_num = True
            pad = [pad, pad]
        self.pads = [-pad[0], -pad[1]]
        if "center" in self._loc:
            self.pads[0] *= -1
            self.pads[1] *= -1
            if "upper" in self._loc or "lower" in self._loc:
                if pad_num:
                    self.pads[0] = 0
                self.pads[1] *= -1
            elif "left" in self._loc or "right" in self._loc:
                if pad_num:
                    self.pads[1] = 0
                self.pads[0] *= -1
        if "left" in self._loc:
            self.pads[0] *= -1
        if "lower" in self._loc:
            self.pads[1] *= -1

    def add_pad(self, delta):
        self.pads[0] += delta[0]
        self.pads[1] += delta[1]

    def sizes(self, ax):
        ax_sizes = (
            ax.get_window_extent()
            .transformed(ax.figure.dpi_scale_trans.inverted())
            .bounds[2:]
        )
        return [
            float(sz[:-1]) / 100 if isinstance(sz, str) else sz / ax_sizes[i]
            for i, sz in enumerate([self._width, self._height])
        ]

    def _size_to_bounds(self, ax):
        sizes = self.sizes(ax)
        origin = [1 - sizes[0], 1 - sizes[1]]
        if "center" in self._loc:
            origin[0] /= 2
            origin[1] /= 2
            if "upper" in self._loc or "lower" in self._loc:
                origin[1] *= 2
            elif "left" in self._loc or "right" in self._loc:
                origin[0] *= 2
        if "left" in self._loc:
            origin[0] = 0
        if "lower" in self._loc:
            origin[1] = 0
        return origin + sizes


def _gen_cax(ax, width=4.0, aspect=7.0, pad=3.0, horiz=False, **kwargs):
    w, h = width / 72, aspect * width / 72
    if horiz:
        cax = _ez_inset(ax, h, w, pad=(0, -w - pad / 72), **kwargs)
    else:
        cax = _ez_inset(ax, w, h, pad=(-w - pad / 72, 0), **kwargs)
    return cax


# TODO: fix colorbar size properly
def nice_colorbar(
    ax: matplotlib.axes.Axes | None = None,
    mappable: matplotlib.cm.ScalarMappable | None = None,
    width: float = 5.0,
    aspect: float = 5.0,
    pad: float = 3.0,
    minmax: bool = False,
    orientation: Literal["vertical", "horizontal"] = "vertical",
    floating=False,
    ticklabels: Sequence[str] | None = None,
    **kwargs: dict,
):
    r"""Creates a colorbar with fixed width and aspect to ensure uniformity of plots.

    Parameters
    ----------
    ax
        The `matplotlib.axes.Axes` instance in which the colorbar is drawn.
    mappable
        The mappable whose colormap and norm will be used.
    width
        The width of the colorbar in points.
    aspect
        aspect ratio of the colorbar.
    pad
        The pad between the colorbar and axes in points.
    minmax
        If `False`, the ticks and the ticklabels will be determined from the keyword
        arguments (the default). If `True`, the minimum and maximum of the colorbar will
        be labeled.
    orientation
        Colorbar orientation.
    **kwargs
        Keyword arguments are passed to `proportional_colorbar`.

    Returns
    -------
    cbar : matplotlib.colorbar.Colorbar
        The created colorbar.

    """

    is_horizontal = orientation == "horizontal"

    if ax is None:
        ax = plt.gca()

    if floating:
        if isinstance(ax, np.ndarray):
            if ax.ndim == 1:
                parent = ax[-1]
            elif ax.ndim == 2:
                parent = ax[0, -1]
            else:
                raise ValueError
        else:
            parent = ax

        cbar = proportional_colorbar(
            mappable=mappable,
            ax=ax,
            cax=_gen_cax(parent, width, aspect, pad, is_horizontal),
            orientation=orientation,
            **kwargs,
        )

    else:
        if np.iterable(ax):
            bbox = mtransforms.Bbox.union(
                [
                    x.get_window_extent().transformed(
                        x.figure.dpi_scale_trans.inverted()
                    )
                    for x in ax.flat
                ]
            )
        else:
            bbox = ax.get_window_extent().transformed(
                ax.figure.dpi_scale_trans.inverted()
            )

        if orientation == "horizontal":
            kwargs["anchor"] = (1, 1)
            kwargs["location"] = "top"
            kwargs["fraction"] = width / (72 * bbox.height)
            kwargs["pad"] = pad / (72 * bbox.height)
            kwargs["shrink"] = width * aspect / (72 * bbox.width)
        else:
            kwargs["anchor"] = (0, 1)
            kwargs["fraction"] = width / (72 * bbox.width)
            kwargs["pad"] = pad / (72 * bbox.width)
            kwargs["shrink"] = width * aspect / (72 * bbox.height)

        cbar = proportional_colorbar(
            mappable=mappable,
            ax=ax,
            aspect=aspect,
            panchor=(0, 1),
            orientation=orientation,
            **kwargs,
        )

    if minmax:
        if is_horizontal:
            cbar.set_ticks(cbar.ax.get_xlim())
        else:
            cbar.set_ticks(cbar.ax.get_ylim())
        cbar.set_ticklabels(("Min", "Max"))
        cbar.ax.tick_params(labelsize="small")

    if ticklabels is not None:
        cbar.set_ticklabels(ticklabels)

    if is_horizontal:
        cbar.ax.set_box_aspect(1 / aspect)
    else:
        cbar.ax.set_box_aspect(aspect)

    return cbar


def flatten_transparency(rgba, background: Sequence[float] | None = None):
    if background is None:
        background = (1, 1, 1)
    original_shape = rgba.shape
    rgba = rgba.reshape(-1, 4)
    rgb = rgba[:, :-1]
    a = rgba[:, -1][:, np.newaxis]
    rgb *= a
    rgb += (1 - a) * background
    return rgb.reshape(original_shape[:-1] + (3,))


def _is_segment_iterable(cmap: mcolors.Colormap) -> bool:
    if not isinstance(cmap, mcolors.LinearSegmentedColormap):
        return False
    if any(callable(cmap._segmentdata[c]) for c in ["red", "green", "blue"]):
        return False
    return True


def combined_cmap(
    cmap1: mcolors.Colormap | str,
    cmap2: mcolors.Colormap | str,
    name: str,
    register: bool = False,
    N=256,
):
    """Stitch two existing colormaps to create a new colormap."""
    if isinstance(cmap1, str):
        cmap1 = matplotlib.colormaps[cmap1]
    if isinstance(cmap2, str):
        cmap2 = matplotlib.colormaps[cmap2]

    if all(_is_segment_iterable(c) for c in (cmap1, cmap2)):
        segnew = {}
        for c in ["red", "green", "blue"]:
            seg1_c, seg2_c = (
                np.asarray(cmap1._segmentdata[c]),
                np.asarray(cmap2._segmentdata[c]),
            )
            seg1_c[:, 0] = seg1_c[:, 0] * 0.5
            seg2_c[:, 0] = seg2_c[:, 0] * 0.5 + 0.5
            segnew[c] = np.r_[seg1_c, seg2_c]
        cmap = mcolors.LinearSegmentedColormap(name=name, segmentdata=segnew, N=N)
    else:
        cmap = mcolors.LinearSegmentedColormap.from_list(
            name=name,
            colors=np.r_[
                cmap1(np.linspace(0, 1, int(N / 2))),
                cmap2(np.linspace(0, 1, int(N / 2))),
            ],
        )
    if register:
        matplotlib.colormaps.register(cmap)
        matplotlib.colormaps.register(cmap.reversed())
    return cmap


def gen_2d_colormap(
    ldat,
    cdat,
    cmap: mcolors.Colormap | str = None,
    colorbar: bool = True,
    *,
    lnorm: plt.Normalize | None = None,
    cnorm: plt.Normalize | None = None,
    background: Sequence[float] | None = None,
    N: int = 256,
):
    if cmap is None:
        cmap = mcolors.LinearSegmentedColormap.from_list(
            "", colors=[[0, 0, 1], [0, 0, 0], [1, 0, 0]], N=N
        )
    elif isinstance(cmap, str):
        cmap = plt.get_cmap(cmap)
    if lnorm is None:
        lnorm = plt.Normalize()
    if cnorm is None:
        cnorm = plt.Normalize()
    if background is None:
        background = (1, 1, 1)

    lnorm.autoscale_None(ldat)
    cnorm.autoscale_None(cdat)

    l_vals = lnorm(ldat)[:, :, np.newaxis]
    img = cmap(cnorm(cdat))[:, :, :-1]
    img *= l_vals
    img += (1 - l_vals) * background

    if colorbar:
        l_vals = lnorm(np.linspace(lnorm.vmin, lnorm.vmax, N))[
            :, np.newaxis, np.newaxis
        ]
        cmap_img = np.repeat(
            cmap(cnorm(np.linspace(cnorm.vmin, cnorm.vmax, N)))[np.newaxis, :], N, 0
        )[:, :, :-1]
        cmap_img *= l_vals
        cmap_img += (1 - l_vals) * background
        return cmap_img, img
    else:
        return None, img


def color_distance(c1, c2):
    # https://www.compuphase.com/cmetric.htm
    R1, G1, B1 = (np.array(mcolors.to_rgb(c1)) * 255).astype(int)
    R2, G2, B2 = (np.array(mcolors.to_rgb(c2)) * 255).astype(int)
    dR2 = (R2 - R1) ** 2
    dG2 = (G2 - G1) ** 2
    dB2 = (B2 - B1) ** 2
    r = 0.5 * (R1 + R2) / 256
    return np.sqrt((2 + r) * dR2 + 4 * dG2 + (2 + 255 / 256 - r) * dB2)


def close_to_white(c):
    c2k = color_distance(c, (0, 0, 0))
    c2w = color_distance(c, (1, 1, 1))
    return bool(c2k > c2w)


def prominent_color(im):
    im_array = im.get_array()
    if im_array is None:
        return mcolors.to_rgba("w")
    hist, edges = np.histogram(np.nan_to_num(im_array), "auto")
    mx = hist.argmax()
    return im.to_rgba(edges[mx : mx + 2].mean())


def image_is_light(im):
    return close_to_white(prominent_color(im))


def axes_textcolor(ax, light="k", dark="w"):
    c = light
    mappable = get_mappable(ax, silent=True)
    if mappable is not None:
        if isinstance(
            mappable, matplotlib.image._ImageBase | matplotlib.collections.QuadMesh
        ):
            if not image_is_light(mappable):
                c = dark
    return c


try:
    combined_cmap("bone_r", "hot", "bonehot", register=True)
except ValueError:
    pass



"""Convenient access to various plotting functions."""

__all__ = [
    "copy_mathtext",
    "fancy_labels",
    "label_subplot_properties",
    "label_subplots",
    "label_subplots_nature",
    "mark_points",
    "mark_points_outside",
    "plot_hv_text",
    "property_label",
    "scale_units",
    "set_titles",
    "set_xlabels",
    "set_ylabels",
    "sizebar",
    "get_bz_edge",
    "plot_hex_bz",
    "InversePowerNorm",
    "TwoSlopePowerNorm",
    "CenteredPowerNorm",
    "TwoSlopeInversePowerNorm",
    "CenteredInversePowerNorm",
    "get_mappable",
    "unify_clim",
    "flatten_transparency",
    "image_is_light",
    "nice_colorbar",
    "proportional_colorbar",
    "fermiline",
    "figwh",
    "autoscale_to",
    "place_inset",
    "plot_array",
    "plot_array_2d",
    "gradient_fill",
    "plot_slices",
]

import numpy as np

from erlab.plotting.annotations import (
    copy_mathtext,
    fancy_labels,
    label_subplot_properties,
    label_subplots,
    label_subplots_nature,
    mark_points,
    mark_points_outside,
    plot_hv_text,
    property_label,
    scale_units,
    set_titles,
    set_xlabels,
    set_ylabels,
    sizebar,
)
from erlab.plotting.bz import get_bz_edge, plot_hex_bz
from erlab.plotting.colors import (
    CenteredInversePowerNorm,
    CenteredPowerNorm,
    InversePowerNorm,
    TwoSlopeInversePowerNorm,
    TwoSlopePowerNorm,
    flatten_transparency,
    get_mappable,
    image_is_light,
    nice_colorbar,
    proportional_colorbar,
    unify_clim,
)
from erlab.plotting.general import (
    autoscale_to,
    fermiline,
    figwh,
    gradient_fill,
    place_inset,
    plot_array,
    plot_array_2d,
    plot_slices,
)


def clean_labels(axes, tick_right=False, *args, **kwargs):
    if axes.ndim == 1:
        axes = axes[None]
    for ax in axes[:-1, :].flat:
        ax.set_xlabel("")
    if tick_right:
        target = axes[:, :-1]
    else:
        target = axes[:, 1:]
    for ax in target.flat:
        ax.set_ylabel("")
    fancy_labels(axes, *args, **kwargs)
    for ax in axes.flat:
        if tick_right:
            ax.yaxis.set_label_position("right")


def integer_ticks(axes):
    if np.iterable(axes):
        for ax in np.asarray(axes, dtype=object):
            integer_ticks(ax)
        return
    axes.set_xticks(
        [
            t
            for t in axes.get_xticks()
            if t.is_integer() and t >= axes.get_xlim()[0] and t <= axes.get_xlim()[1]
        ]
    )
    axes.set_yticks(
        [
            t
            for t in axes.get_yticks()
            if t.is_integer() and t >= axes.get_ylim()[0] and t <= axes.get_ylim()[1]
        ]
    )



__all__ = ["goldtool"]

import os
import time
from typing import TYPE_CHECKING

import joblib
import numpy as np
import pyqtgraph as pg
import varname
import xarray as xr
from qtpy import QtCore, QtWidgets

import erlab.analysis
from erlab.interactive.imagetool import ImageTool
from erlab.interactive.utilities import (
    AnalysisWindow,
    ParameterGroup,
    ROIControls,
    gen_function_code,
)
from erlab.parallel import joblib_progress_qt

if TYPE_CHECKING:
    import lmfit
    import scipy.interpolate

LMFIT_METHODS = [
    "leastsq",
    "least_squares",
    "differential_evolution",
    # "brute",
    # "basinhopping",
    # "ampgo",
    # "nelder",
    # "lbfgsb",
    # "powell",
    "cg",
    # "newton",
    # "cobyla",
    # "bfgs",
    # "tnc",
    # "trust-ncg",
    # "trust-exact",
    # "trust-krylov",
    # "trust-constr",
    # "dogleg",
    # "slsqp",
    "emcee",
    # "shgo",
    # "dual_annealing",
]


class EdgeFitter(QtCore.QThread):
    sigIterated = QtCore.Signal(int)
    sigFinished = QtCore.Signal()

    def set_params(self, data, x0, y0, x1, y1, params):
        self.data = data
        self.x_range: tuple[float, float] = (x0, x1)
        self.y_range: tuple[float, float] = (y0, y1)
        self.params = params
        self.parallel_obj = joblib.Parallel(
            n_jobs=self.params["# CPU"], max_nbytes=None
        )

    @QtCore.Slot()
    def abort_fit(self):
        self.parallel_obj._aborting = True
        self.parallel_obj._exception = True

    def run(self):
        self.sigIterated.emit(0)
        with joblib_progress_qt(self.sigIterated) as _:
            self.edge_center, self.edge_stderr = erlab.analysis.gold.edge(
                gold=self.data,
                angle_range=self.x_range,
                eV_range=self.y_range,
                bin_size=(self.params["Bin x"], self.params["Bin y"]),
                temp=self.params["T (K)"],
                vary_temp=not self.params["Fix T"],
                fast=self.params["Fast"],
                method=self.params["Method"],
                scale_covar=self.params["Scale cov"],
                progress=False,
                parallel_obj=self.parallel_obj,
            )
        self.sigFinished.emit()


class GoldTool(AnalysisWindow):
    """Interactive gold edge fitting.

    Parameters
    ----------
    data
        The data to perform Fermi edge fitting on.
    data_corr
        The data to correct with the edge. Defaults to `data`.
    data_name
        Name of the data used in generating the code snipped copied to the clipboard.
        Overrides automatic detection.
    **kwargs
        Arguments passed onto `erlab.interactive.utilities.AnalysisWindow`.

    Signals
    -------
    sigProgressUpdated(int)
        Signal used to update the progress bar.
    sigAbortFitting()
        Signal used to abort the fitting, emitted when the cancel button is clicked.

    """

    sigProgressUpdated = QtCore.Signal(int)  #: :meta private:
    sigAbortFitting = QtCore.Signal()  #: :meta private:

    def __init__(
        self,
        data: xr.DataArray,
        data_corr: xr.DataArray | None = None,
        *,
        data_name: str | None = None,
        **kwargs: dict,
    ):
        super().__init__(
            data,
            link="x",
            layout="horizontal",
            orientation="vertical",
            num_ax=3,
            **kwargs,
        )

        self._argnames = {}
        if data_name is None:
            try:
                self._argnames["data"] = varname.argname(
                    "data", func=self.__init__, vars_only=False
                )
            except varname.VarnameRetrievingError:
                self._argnames["data"] = "gold"
        else:
            self._argnames["data"] = data_name

        if data_corr is not None:
            try:
                self._argnames["data_corr"] = varname.argname(
                    "data_corr", func=self.__init__, vars_only=False
                )
            except varname.VarnameRetrievingError:
                self._argnames["data_corr"] = "data_corr"

        self.data_corr = data_corr

        self.axes[1].setVisible(False)
        self.hists[1].setVisible(False)
        self.axes[2].setVisible(False)
        self.hists[2].setVisible(False)

        try:
            temp = float(self.data.attrs["temp_sample"])
        except KeyError:
            temp = 30.0

        self.params_roi = ROIControls(self.add_roi(0))
        self.params_edge = ParameterGroup(
            **{
                "T (K)": {"qwtype": "dblspin", "value": temp, "range": (0.0, 400.0)},
                "Fix T": {"qwtype": "chkbox", "checked": True},
                "Bin x": {"qwtype": "spin", "value": 1, "minimum": 1},
                "Bin y": {"qwtype": "spin", "value": 1, "minimum": 1},
                "Fast": {"qwtype": "chkbox", "checked": False},
                "Method": {"qwtype": "combobox", "items": LMFIT_METHODS},
                "Scale cov": {"qwtype": "chkbox", "checked": True},
                "# CPU": {
                    "qwtype": "spin",
                    "value": os.cpu_count(),
                    "minimum": 1,
                    "maximum": os.cpu_count(),
                },
                "go": {
                    "qwtype": "pushbtn",
                    "showlabel": False,
                    "text": "Go",
                    "clicked": self.perform_edge_fit,
                },
            }
        )

        self.params_edge.widgets["Fast"].stateChanged.connect(self._toggle_fast)

        self.params_poly = ParameterGroup(
            **{
                "Degree": {"qwtype": "spin", "value": 4, "range": (1, 20)},
                "Method": {"qwtype": "combobox", "items": LMFIT_METHODS},
                "Scale cov": {"qwtype": "chkbox", "checked": True},
                "Residuals": {"qwtype": "chkbox", "checked": False},
                "Corrected": {"qwtype": "chkbox", "checked": False},
                "Shift coords": {"qwtype": "chkbox", "checked": True},
                "itool": {
                    "qwtype": "pushbtn",
                    "notrack": True,
                    "showlabel": False,
                    "text": "Open in ImageTool",
                    "clicked": self.open_itool,
                },
                "copy": {
                    "qwtype": "pushbtn",
                    "notrack": True,
                    "showlabel": False,
                    "text": "Copy to clipboard",
                    "clicked": lambda: self.gen_code("poly"),
                },
            }
        )

        self.params_spl = ParameterGroup(
            **{
                "Auto": {"qwtype": "chkbox", "checked": True},
                "lambda": {
                    "qwtype": "dblspin",
                    "minimum": 0,
                    "maximum": 10000,
                    "singleStep": 0.001,
                    "decimals": 4,
                },
                "Residuals": {"qwtype": "chkbox", "checked": False},
                "Corrected": {"qwtype": "chkbox", "checked": False},
                "Shift coords": {"qwtype": "chkbox", "checked": True},
                "itool": {
                    "qwtype": "pushbtn",
                    "notrack": True,
                    "showlabel": False,
                    "text": "Open in ImageTool",
                    "clicked": self.open_itool,
                },
                "copy": {
                    "qwtype": "pushbtn",
                    "notrack": True,
                    "showlabel": False,
                    "text": "Copy to clipboard",
                    "clicked": lambda: self.gen_code("spl"),
                },
            }
        )
        self.params_spl.widgets["lambda"].setDisabled(
            self.params_spl.widgets["Auto"].checkState() == QtCore.Qt.CheckState.Checked
        )
        self.params_spl.widgets["Auto"].toggled.connect(
            lambda _: self.params_spl.widgets["lambda"].setDisabled(
                self.params_spl.widgets["Auto"].checkState()
                == QtCore.Qt.CheckState.Checked
            )
        )

        self.controls.addWidget(self.params_roi)
        self.controls.addWidget(self.params_edge)

        self.params_tab = QtWidgets.QTabWidget()
        self.params_tab.addTab(self.params_poly, "Polynomial")
        self.params_tab.addTab(self.params_spl, "Spline")
        self.params_tab.currentChanged.connect(
            lambda i: self.perform_fit(("poly", "spl")[i])
        )
        self.controls.addWidget(self.params_tab)

        self.params_poly.setDisabled(True)
        self.params_spl.setDisabled(True)
        self.params_tab.setDisabled(True)

        self.controlgroup.setSizePolicy(
            QtWidgets.QSizePolicy.Policy.Maximum, QtWidgets.QSizePolicy.Policy.Preferred
        )

        self.scatterplots = [
            pg.ScatterPlotItem(
                size=3,
                pen=pg.mkPen(None),
                brush=pg.mkBrush(255, 0, 0, 200),
                pxMode=True,
            )
            for _ in range(2)
        ]

        self.errorbars = [pg.ErrorBarItem(pen=pg.mkPen(color="red")) for _ in range(2)]
        self.polycurves = [pg.PlotDataItem() for _ in range(2)]
        for i in range(2):
            self.axes[i].addItem(self.scatterplots[i])
            self.axes[i].addItem(self.errorbars[i])
            self.axes[i].addItem(self.polycurves[i])
        self.params_poly.sigParameterChanged.connect(lambda: self.perform_fit("poly"))
        self.params_spl.sigParameterChanged.connect(lambda: self.perform_fit("spl"))

        self.axes[0].disableAutoRange()

        # Setup time calculation
        self.start_time: float | None = None
        self.step_times: list[float] = []

        # Setup progress bar
        self.progress = QtWidgets.QProgressDialog(
            labelText="Fitting...",
            minimum=0,
            parent=self,
            minimumDuration=0,
            windowModality=QtCore.Qt.WindowModal,
        )
        self.pbar = QtWidgets.QProgressBar()
        self.progress.setBar(self.pbar)
        self.progress.setFixedSize(self.progress.size())
        self.progress.setCancelButtonText("Abort!")
        self.progress.canceled.disconnect(self.progress.cancel)  # don't auto close
        self.progress.canceled.connect(self.abort_fit)
        self.progress.setAutoReset(False)
        self.progress.cancel()

        # Setup fitter thread
        # This allows the GUI to remain responsive during fitting so it can be aborted
        self.fitter = EdgeFitter()
        self.fitter.sigIterated.connect(self.iterated)
        self.fitter.sigFinished.connect(self.post_fit)
        self.sigAbortFitting.connect(self.fitter.abort_fit)

        # Resize roi to data bounds
        eV_span = self.data.eV.values[-1] - self.data.eV.values[0]
        ang_span = self.data.alpha.values[-1] - self.data.alpha.values[0]
        x1 = self.data.alpha.values.mean() + ang_span * 0.45
        x0 = self.data.alpha.values.mean() - ang_span * 0.45
        y1 = self.data.eV.values[-1] - eV_span * 0.015
        y0 = y1 - eV_span * 0.3
        self.params_roi.modify_roi(x0, y0, x1, y1)

        # Initialize fit result
        self.result: scipy.interpolate.BSpline | lmfit.model.ModelResult | None = None

        self.__post_init__(execute=True)

    def _toggle_fast(self):
        self.params_edge.widgets["T (K)"].setDisabled(self.params_edge.values["Fast"])
        self.params_edge.widgets["Fix T"].setDisabled(self.params_edge.values["Fast"])

    def iterated(self, n: int):
        self.step_times.append(time.perf_counter() - self.start_time)
        self.progress.setValue(n)

        deltas = np.diff(self.step_times)
        timeleft = (self.progress.maximum() - (n - 1)) * np.mean(deltas)

        # timeleft: str = humanize.precisedelta(datetime.timedelta(seconds=timeleft))
        # steptime: str = humanize.precisedelta(datetime.timedelta(seconds=steptime))

        self.progress.setLabelText(f"{round(timeleft)} seconds left...")
        self.pbar.setFormat(f"{n}/{self.progress.maximum()} finished")

    @QtCore.Slot()
    def perform_edge_fit(self):
        self.start_time = time.perf_counter()
        self.step_times: list[float] = [0.0]

        self.progress.setVisible(True)
        self.params_roi.draw_button.setChecked(False)
        x0, y0, x1, y1 = (np.round(x, 3) for x in self.params_roi.roi_limits)
        params = self.params_edge.values
        n_total = len(
            self.data.alpha.coarsen(alpha=params["Bin x"], boundary="trim")
            .mean()
            .sel(alpha=slice(x0, x1))
        )
        self.progress.setMaximum(n_total)
        self.fitter.set_params(self.data, x0, y0, x1, y1, params)
        self.fitter.start()

    @QtCore.Slot()
    def abort_fit(self):
        self.sigAbortFitting.emit()

    def closeEvent(self, event):
        super().closeEvent(event)

    @QtCore.Slot()
    def post_fit(self):
        self.progress.reset()
        self.edge_center, self.edge_stderr = (
            self.fitter.edge_center,
            self.fitter.edge_stderr,
        )

        xval = self.edge_center.alpha.values
        yval = self.edge_center.values
        for i in range(2):
            self.scatterplots[i].setData(x=xval, y=yval)
            self.errorbars[i].setData(x=xval, y=yval, height=self.edge_stderr.values)

        self.params_poly.setDisabled(False)
        self.params_spl.setDisabled(False)
        self.params_tab.setDisabled(False)
        self.perform_fit("poly")

    def perform_fit(self, mode="poly"):
        match mode:
            case "poly":
                edgefunc = self._perform_poly_fit()
                params = self.params_poly.values
            case "spl":
                edgefunc = self._perform_spline_fit()
                params = self.params_spl.values
        for i in range(2):
            xval = self.data.alpha.values
            if i == 1 and params["Residuals"]:
                yval = np.zeros_like(xval)
            else:
                yval = edgefunc(xval)
            self.polycurves[i].setData(x=xval, y=yval)

        xval = self.edge_center.alpha.values
        if params["Residuals"]:
            yval = edgefunc(xval) - self.edge_center.values
        else:
            yval = self.edge_center.values
        self.errorbars[1].setData(x=xval, y=yval)
        self.scatterplots[1].setData(x=xval, y=yval, height=self.edge_stderr)

        self.aw.axes[1].setVisible(True)
        self.aw.images[-1].setDataArray(self.corrected)
        self.aw.axes[2].setVisible(params["Corrected"])
        self.aw.hists[2].setVisible(params["Corrected"])

    def _perform_poly_fit(self):
        params = self.params_poly.values
        self.result = erlab.analysis.gold.poly_from_edge(
            center=self.edge_center,
            weights=1 / self.edge_stderr,
            degree=params["Degree"],
            method=params["Method"],
            scale_covar=params["Scale cov"],
        )
        if self.data_corr is None:
            target = self.data
        else:
            target = self.data_corr
        self.corrected = erlab.analysis.correct_with_edge(
            target, self.result, plot=False, shift_coords=params["Shift coords"]
        )
        return lambda x: self.result.eval(self.result.params, x=x)

    def _perform_spline_fit(self):
        params = self.params_spl.values
        if params["Auto"]:
            params["lambda"] = None
        self.result = erlab.analysis.gold.spline_from_edge(
            center=self.edge_center,
            weights=np.asarray(1 / self.edge_stderr),
            lam=params["lambda"],
        )

        if self.data_corr is None:
            target = self.data
        else:
            target = self.data_corr
        self.corrected = erlab.analysis.correct_with_edge(
            target, self.result, plot=False, shift_coords=params["Shift coords"]
        )
        return self.result

    def open_itool(self):
        self.itool = ImageTool(self.corrected)
        self.itool.show()

    def gen_code(self, mode: str):
        p0 = self.params_edge.values
        match mode:
            case "poly":
                p1 = self.params_poly.values
            case "spl":
                p1 = self.params_spl.values
        x0, y0, x1, y1 = (np.round(x, 3) for x in self.params_roi.roi_limits)

        arg_dict = {
            "angle_range": (x0, x1),
            "eV_range": (y0, y1),
            "bin_size": (p0["Bin x"], p0["Bin y"]),
            "temp": p0["T (K)"],
            "method": p0["Method"],
        }

        match mode:
            case "poly":
                fname = "poly"
                arg_dict["degree"] = p1["Degree"]
            case "spl":
                fname = "spline"
                if p1["Auto"]:
                    arg_dict["lam"] = None
                else:
                    arg_dict["lam"] = p1["lambda"]

        if p0["Fast"]:
            arg_dict["fast"] = True
            del arg_dict["temp"]
        elif not p0["Fix T"]:
            arg_dict["vary_temp"] = True

        if not p0["Scale cov"]:
            arg_dict["scale_covar_edge"] = False

        if mode == "poly":
            if not p1["Scale cov"]:
                arg_dict["scale_covar"] = False

        if self.data_corr is None:
            gen_function_code(
                copy=True,
                **{
                    f"modelresult = era.gold.{fname}": [
                        f"|{self._argnames['data']}|",
                        arg_dict,
                    ]
                },
            )
        else:
            arg_dict["correct"] = True
            gen_function_code(
                copy=True,
                **{
                    f"modelresult = era.gold.{fname}": [
                        f"|{self._argnames['data']}|",
                        arg_dict,
                    ],
                    "corrected = era.correct_with_edge": [
                        f"|{self._argnames['data_corr']}|",
                        "|modelresult|",
                        {"shift_coords": p1["Shift coords"]},
                    ],
                },
            )


def goldtool(
    data: xr.DataArray,
    data_corr: xr.DataArray | None = None,
    *,
    data_name: str | None = None,
    **kwargs: dict,
) -> GoldTool:
    """Interactive gold edge fitting.

    Parameters
    ----------
    data
        The data to perform Fermi edge fitting on.
    data_corr
        The data to correct with the edge. Defaults to `data`.
    data_name
        Name of the data used in generating the code snipped copied to the clipboard.
        Overrides automatic detection.
    **kwargs
        Arguments passed onto `erlab.interactive.utilities.AnalysisWindow`.
    """
    if data_name is None:
        try:
            data_name = varname.argname("data", func=goldtool, vars_only=False)
        except varname.VarnameRetrievingError:
            data_name = "data"
    return GoldTool(data, data_corr, data_name=data_name, **kwargs)



"""Fast linear interpolation."""

__all__ = ["FastInterpolator", "interpn"]

import math
import warnings
from collections.abc import Sequence

import numba
import numpy as np
import scipy.interpolate
import xarray as xr
import xarray.core.missing


@numba.njit(nogil=True, inline="always")
def _do_interp1(x, v0, v1):
    return v0 * (1 - x) + v1 * x


@numba.njit(nogil=True, inline="always")
def _do_interp2(x, y, v0, v1, v2, v3):
    return _do_interp1(y, _do_interp1(x, v0, v1), _do_interp1(x, v2, v3))


@numba.njit(nogil=True, inline="always")
def _do_interp3(x, y, z, v0, v1, v2, v3, v4, v5, v6, v7):
    return _do_interp1(
        z, _do_interp2(x, y, v0, v1, v2, v3), _do_interp2(x, y, v4, v5, v6, v7)
    )


# @numba.njit
# def _do_interp2(x, y, v0, v1, v2, v3):
#     return _do_interp1(y, v0 * (1 - x) + v1 * x, v2 * (1 - x) + v3 * x)


# @numba.njit
# def _do_interp3(x, y, z, v0, v1, v2, v3, v4, v5, v6, v7):
#     return (
#         _do_interp2(x, y, v0, v1, v2, v3) * (1 - z)
#         + _do_interp2(x, y, v4, v5, v6, v7) * z
#     )


@numba.njit(nogil=True, inline="always")
def _calc_interp2(values, v0, v1):
    i0, i1 = math.floor(v0), math.floor(v1)
    n0, n1 = values.shape
    j0, j1 = min(i0 + 1, n0 - 1), min(i1 + 1, n1 - 1)
    return _do_interp2(
        v0 - i0,
        v1 - i1,
        values[i0, i1],
        values[j0, i1],
        values[i0, j1],
        values[j0, j1],
    )


@numba.njit(nogil=True, inline="always")
def _calc_interp3(values, v0, v1, v2):
    i0, i1, i2 = math.floor(v0), math.floor(v1), math.floor(v2)
    n0, n1, n2 = values.shape
    j0, j1, j2 = min(i0 + 1, n0 - 1), min(i1 + 1, n1 - 1), min(i2 + 1, n2 - 1)
    return _do_interp3(
        v0 - i0,
        v1 - i1,
        v2 - i2,
        values[i0, i1, i2],
        values[j0, i1, i2],
        values[i0, j1, i2],
        values[j0, j1, i2],
        values[i0, i1, j2],
        values[j0, i1, j2],
        values[i0, j1, j2],
        values[j0, j1, j2],
    )


@numba.njit(nogil=True, inline="always")
def _val2ind(val, coord):
    if val > coord[-1] or val < coord[0]:
        return np.nan
    else:
        return np.divide(val - coord[0], coord[1] - coord[0])


@numba.njit(nogil=True, parallel=True)
def _interp2(x, y, values, xc, yc, fill_value=np.nan):
    n = len(xc)

    arr_new = np.empty(n, values.dtype)

    for m in numba.prange(n):
        v0, v1 = _val2ind(xc[m], x), _val2ind(yc[m], y)

        if np.isnan(v0) or np.isnan(v1):
            arr_new[m] = fill_value
        else:
            arr_new[m] = _calc_interp2(values, v0, v1)
    return arr_new


@numba.njit(nogil=True, parallel=True)
def _interp3(x, y, z, values, xc, yc, zc, fill_value=np.nan):
    n = len(xc)

    arr_new = np.empty(n, values.dtype)

    for m in numba.prange(n):
        v0, v1, v2 = _val2ind(xc[m], x), _val2ind(yc[m], y), _val2ind(zc[m], z)

        if np.isnan(v0) or np.isnan(v1) or np.isnan(v2):
            arr_new[m] = fill_value
        else:
            arr_new[m] = _calc_interp3(values, v0, v1, v2)
    return arr_new


def _get_interp_func(ndim):
    if ndim == 3:
        return _interp3
    elif ndim == 2:
        return _interp2
    else:
        raise ValueError("Fast interpolation only supported for 2D or 3D")


def _check_even(arr):
    dif = np.diff(arr)
    return np.allclose(dif, dif[0])


class FastInterpolator(scipy.interpolate.RegularGridInterpolator):
    """Fast linear multidimensional interpolation on evenly spaced coordinates.

    This is an extension from `scipy.interpolate.RegularGridInterpolator` with vast
    performance improvements and integration with `xarray`.

    The input arguments are identical to `scipy.interpolate.RegularGridInterpolator`
    except `bounds_error`, which is set to `False` by default.

    Performance improvements are enabled for 2D and 3D linear interpolation on uniformly
    spaced coordinates with extrapolation disabled. Otherwise,
    `scipy.interpolate.RegularGridInterpolator` is called. See below for more
    information.

    Note
    ----
    Parallel acceleration is only applied when all of the following are true.

    * `method` is ``"linear"``.
    * Coordinates along all dimensions are evenly spaced.
    * Values are 2D or 3D.
    * Extrapolation is disabled, i.e., `fill_value` is not `None`.
    * The dimension of coordinates `xi` matches the number of dimensions of the values.
      Also, each coordinate array in `xi` must have the same shape.

    See Also
    --------
    interpn : a convenience function which wraps `FastInterpolator`

    """

    def __init__(
        self,
        points,
        values,
        method="linear",
        bounds_error=False,
        fill_value=np.nan,
    ):
        super().__init__(
            points,
            values,
            method=method,
            bounds_error=bounds_error,
            fill_value=fill_value,
        )

        self.uneven_dims = tuple(
            i for i, g in enumerate(self.grid) if not _check_even(g)
        )

    @classmethod
    def from_xarray(
        cls,
        data: xr.DataArray,
        method="linear",
        bounds_error: bool = False,
        fill_value: float | None = np.nan,
    ):
        """Construct an interpolator from a `xarray.DataArray`.

        Parameters
        ----------
        data
            The source `xarray.DataArray`.
        method
            The method of interpolation to perform.
        fill_value
            The value to use for points outside of the interpolation domain.

        """
        return cls(
            tuple(data[dim].values for dim in data.dims),
            data.values,
            method=method,
            bounds_error=bounds_error,
            fill_value=fill_value,
        )

    def __call__(self, xi, method=None):
        is_linear: bool = method == "linear" or self.method == "linear"
        nd_supported: bool = self.values.ndim in (2, 3)
        no_extrap: bool = self.fill_value is not None

        if (len(self.uneven_dims) == 0) and is_linear and nd_supported and no_extrap:
            if isinstance(xi, np.ndarray):
                xi = tuple(xi.take(i, axis=-1) for i in range(xi.shape[-1]))

            xi_shapes = [x.shape for x in xi]
            if not all(s == xi_shapes[0] for s in xi_shapes):
                warnings.warn(
                    "Not all coordinate arrays have the same shape, "
                    "falling back to scipy.",
                    RuntimeWarning,
                    stacklevel=1,
                )
            elif len(xi) != self.values.ndim:
                warnings.warn(
                    f"Number of input dimensions ({len(xi)}) does not match "
                    "the input data dimensions, "
                    "falling back to scipy.",
                    RuntimeWarning,
                    stacklevel=1,
                )
            else:
                result = _get_interp_func(self.values.ndim)(
                    *self.grid,
                    self.values,
                    *(c.ravel() for c in xi),
                    fill_value=self.fill_value,
                ).reshape(xi[0].shape + self.values.shape[self.values.ndim :])
                return result

        if (len(self.uneven_dims) != 0) and is_linear:
            warnings.warn(
                f"Dimension(s) {self.uneven_dims} are not uniform, "
                "falling back to scipy.",
                RuntimeWarning,
                stacklevel=1,
            )
        return super().__call__(xi, method)


def interpn(
    points: Sequence[np.ndarray],
    values: np.ndarray,
    xi: Sequence[np.ndarray] | np.ndarray,
    method: str = "linear",
    bounds_error: bool = False,
    fill_value: float | None = np.nan,
):
    """Multidimensional interpolation on evenly spaced coordinates.

    This can be used as a drop-in replacement for `scipy.interpolate.interpn`.
    Performance optimization is applied in some special cases, documented in
    `FastInterpolator`.

    Parameters
    ----------
    points
        The points defining the regular grid in `n` dimensions. The points in
        each dimension (i.e. every element of the points tuple) must be strictly
        ascending.
    values
        The data on the regular grid in `n` dimensions.
    xi
        The coordinates to sample the gridded data at. In addition to the
        scipy-compatible syntax, a tuple of coordinates is also acceptable.
    method
        The method of interpolation to perform.
    fill_value
        The value to use for points outside of the interpolation domain.

    Returns
    -------
    values_x : numpy.ndarray
        Interpolated values at input coordinates.


    Note
    ----
    This optimized version of linear interpolation can be used with the `xarray`
    interpolation methods `xarray.Dataset.interp` and `xarray.DataArray.interp` by
    supplying ``method="linearfast"``. Note that the fallback to `scipy` will be silent
    except when a non-uniform dimension is found, when a warning will be issued.

    """
    interp = FastInterpolator(
        points, values, method=method, bounds_error=bounds_error, fill_value=fill_value
    )
    return interp(xi)


_get_interpolator_nd_original = xarray.core.missing._get_interpolator_nd
_get_interpolator_original = xarray.core.missing._get_interpolator


def _get_interpolator_fast(method, **kwargs):
    if method == "linearfast":
        method = "linear"
    return _get_interpolator_original(method, **kwargs)


def _get_interpolator_nd_fast(method, **kwargs):
    if method == "linearfast":
        return interpn, kwargs
    else:
        return _get_interpolator_nd_original(method, **kwargs)


xarray.core.missing._get_interpolator = _get_interpolator_fast
xarray.core.missing._get_interpolator_nd = _get_interpolator_nd_fast
